/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.constants.BindingSoapVersion;
import org.ow2.petals.component.framework.constants.SoapEnvelope;
import org.ow2.petals.component.framework.constants.SoapFaultCode;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;


/**
 * This class allows to handle a JBI Fault as a SOAP Fault
 *
 * @author nsalatge - EBM Websourcing
 * @since 4.0
 *
 */
public class SOAPFault {

    private final static String PREFIX_ENVELOPE = "env";

    private Fault jbiFault;

    private List<Element> soapFaults;
    
    private final static ThreadLocal<XMLOutputFactory> xofThreadLocal = new ThreadLocal<XMLOutputFactory>() {

        @Override
        protected XMLOutputFactory initialValue() {
            return XMLOutputFactory.newInstance();
        }
    };

    /**
     * Soap envelope constant
     * @author Mathieu CARROLLE
     * @deprecated Use {@link SoapEnvelope} instead
     */
    @Deprecated
    public enum SoapEnvelopeConstants {
        SOAP11_NS("http://schemas.xmlsoap.org/soap/envelope/"),
        SOAP12_NS("http://www.w3.org/2003/05/soap-envelope");
        
        /**
         * 
         * @param pattern
         * @return
         */
        public static SoapEnvelopeConstants valueOf(final URI pattern) {
            SoapEnvelopeConstants result = null;
            if (pattern != null) {
                for (SoapEnvelopeConstants soap : SoapEnvelopeConstants.values()) {
                    if (soap.nameSpace.equals(pattern.toString())) {
                        result = soap;
                    }
                }
            }
            return result;
        }
        
        private final String nameSpace;
        private final URI soapURI;
        
        private SoapEnvelopeConstants(final String nameSpace) {
            this.nameSpace = nameSpace;
            try {
                this.soapURI = new URI(nameSpace);
            } catch (final URISyntaxException e) {
                throw new Error("Unexpected Error in URI namespace syntax", e); // NOPMD
            }
        }
        
        /**
         * 
         * @return
         */
        public URI value() {
            return this.soapURI;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(mep)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param mep
         * @return
         */
        public boolean equals(final URI mep) {
            return toString().equals(mep.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        @Override
        public String toString() {
            return this.nameSpace;
        }
    }
    
    /**
     * SOAP Version Constants for the Message Exchange Patterns.
     * @deprecated Use {@link BindingSoapVersion}
     */
    @Deprecated
    public enum BindingSoapVersionConstants {
        SOAP11_NS("http://schemas.xmlsoap.org/wsdl/soap/"),
        SOAP12_NS("http://schemas.xmlsoap.org/wsdl/soap12/");
        /**
         * 
         * @param pattern
         * @return
         */
        public static BindingSoapVersionConstants valueOf(final URI pattern) {
            BindingSoapVersionConstants result = null;
            if (pattern != null) {
                for (BindingSoapVersionConstants soap : BindingSoapVersionConstants.values()) {
                    if (soap.nameSpace.equals(pattern.toString())) {
                        result = soap;
                    }
                }
            }
            return result;
        }

        private final String nameSpace;
        private final URI soapURI;

        /**
         * Creates a new instance of {@link MEPConstants}
         * 
         * @param nameSpace
         */
        private BindingSoapVersionConstants(final String nameSpace) {
            this.nameSpace = nameSpace;
            try {
                this.soapURI = new URI(nameSpace);
            } catch (final URISyntaxException e) {
                throw new Error("Unexpected Error in URI namespace syntax", e); // NOPMD
            }
        }

        /**
         * 
         * @return
         */
        public URI value() {
            return this.soapURI;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(mep)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param mep
         * @return
         */
        public boolean equals(URI mep) {
            return toString().equals(mep.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        public String toString() {
            return this.nameSpace;
        }
    }


    /**
     * SOAP Fault Code Constants for the Message Exchange Patterns.
     * 
     * @deprecated Use {@link SoapFaultCode} instead
     */
    @Deprecated
    public enum SoapFaultCodeConstants {
        VERSION_MISMATCH("VersionMismatch"),
        MUST_UNDERSTAND("MustUnderstand"),
        DATA_ENCODING_UNKNOWN("DataEncodingUnknown"),
        SENDER("Sender"),
        RECEIVER("Receiver"),
        CLIENT("Client"),
        SERVER("Server");



        private final String value;

        /**
         * Creates a new instance of {@link FeatureConstants}
         * 
         * @param value
         */
        private SoapFaultCodeConstants(final String value) {
            this.value = value;
        }

        /**
         * 
         * @return
         */
        public String value() {
            return value;
        }

        /**
         * Please use this equals method instead of using :<code>
         * value().equals(value)
         * </code>
         * which is almost 10 times slower...
         * 
         * @param value
         * @return
         */
        public boolean equals(String value) {
            return toString().equals(value.toString());
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Enum#toString()
         */
        public String toString() {
            return this.value;
        }
    }

    public SOAPFault(final Fault jbiFault) throws PEtALSCDKException {
        this.jbiFault = jbiFault;
        this.soapFaults = new ArrayList<Element>();

        Document jbi = SourceHelper.createDocument(this.jbiFault.getContent());

        NodeList details = null;
        if(jbi.getDocumentElement().getNamespaceURI().equals(SoapEnvelope.SOAP11_NS.toString())) {
            details = jbi.getDocumentElement().getElementsByTagName("detail");
        } else if(jbi.getDocumentElement().getNamespaceURI().equals(SoapEnvelope.SOAP12_NS.toString())) {
            details = jbi.getDocumentElement().getElementsByTagName("Detail");
        }

        if((details != null)&&(details.getLength() > 0)) {
            for (int i = 0; i < details.getLength(); i++) {
                if(details.item(i).getNodeType() == Node.ELEMENT_NODE) {
                    soapFaults.add((Element) details.item(i));
                }
            }
        }
    }

    public List<Element> getSoapFaults() {
        return this.soapFaults;
    }

    /**
     * @Deprecated Use {@link SOAPFault#createSOAP11Fault(SoapFaultCode, String, List)} instead.
     * @param code
     * @param reason
     * @param details
     * @return
     * @throws PEtALSCDKException
     */
    @Deprecated
    public static Source createSOAP11Fault(SoapFaultCodeConstants code, String reason, List<Element> details) throws PEtALSCDKException {
        return SOAPFault.createFault(BindingSoapVersion.SOAP11_NS, SoapFaultCode.valueOf(code.name()), reason, details);
    }
    
    public static Source createSOAP11Fault(SoapFaultCode code, String reason, List<Element> details) throws PEtALSCDKException {
        return SOAPFault.createFault(BindingSoapVersion.SOAP11_NS, code, reason, details);
    }
    
    /**
     * @Deprecated Use {@link SOAPFault#createSOAP12Fault(SoapFaultCode, String, List)} instead.
     * @param code
     * @param reason
     * @param details
     * @return
     * @throws PEtALSCDKException
     */
    @Deprecated
    public static Source createSOAP12Fault(SoapFaultCodeConstants code, String reason, List<Element> details) throws PEtALSCDKException {
        return SOAPFault.createFault(BindingSoapVersion.SOAP12_NS, SoapFaultCode.valueOf(code.name()), reason, details);
    }
    
    public static Source createSOAP12Fault(SoapFaultCode code, String reason, List<Element> details) throws PEtALSCDKException {
        return SOAPFault.createFault(BindingSoapVersion.SOAP12_NS, code, reason, details);
    }
    
    private static Source createFault(final BindingSoapVersion version, final SoapFaultCode code, final String reason, final List<Element> details) throws PEtALSCDKException {
        
        if (code.equals(SoapFaultCode.DATA_ENCODING_UNKNOWN) && version.equals(BindingSoapVersion.SOAP11_NS)) {
            throw new PEtALSCDKException("Fault code unknown [" + code.value() + "] in this soap version: " + version);
        }
        
        try {
            final ByteArrayOutputStream baos = new ByteArrayOutputStream(2 * 1024);
            final XMLStreamWriter xsw = xofThreadLocal.get().createXMLStreamWriter(baos);
            
            xsw.writeStartDocument();
            
            SoapFaultCode correctCode = code; 
            if(version.equals(BindingSoapVersion.SOAP11_NS)) {
                xsw.setPrefix(SOAPFault.PREFIX_ENVELOPE, SoapEnvelope.SOAP11_NS.value().toString());
                xsw.writeStartElement(SoapEnvelope.SOAP11_NS.value().toString(), "Envelope");
                xsw.writeNamespace(SOAPFault.PREFIX_ENVELOPE, SoapEnvelope.SOAP11_NS.value().toString());
                xsw.writeStartElement(SoapEnvelope.SOAP11_NS.value().toString(), "Body");
                xsw.writeStartElement(SoapEnvelope.SOAP11_NS.value().toString(), "Fault");
                
                if( code.equals(SoapFaultCode.RECEIVER)) {
                    correctCode = SoapFaultCode.SERVER;
                } else if (code.equals(SoapFaultCode.SENDER)) {
                    correctCode = SoapFaultCode.CLIENT;
                } 
            } else if(version.equals(BindingSoapVersion.SOAP12_NS)) {
                xsw.setPrefix(SOAPFault.PREFIX_ENVELOPE, SoapEnvelope.SOAP12_NS.value().toString());
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Envelope");
                xsw.writeNamespace(SOAPFault.PREFIX_ENVELOPE, SoapEnvelope.SOAP12_NS.value().toString());
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Body");
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Fault");
                
                if (code.equals(SoapFaultCode.CLIENT)) {
                    correctCode = SoapFaultCode.SENDER;
                } else if (code.equals(SoapFaultCode.SERVER)) {
                    correctCode = SoapFaultCode.RECEIVER;
                } 
            } else {
                throw new PEtALSCDKException("Fault unknown");
            }

            // create fault code
            if (version.equals(BindingSoapVersion.SOAP12_NS)) {
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Code");
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Value");
                final String prefixEnv = xsw.getPrefix(SoapEnvelope.SOAP12_NS.value().toString());
                xsw.writeCharacters(
                        (prefixEnv == null || prefixEnv.length() == 0 ? "" : prefixEnv + ":" )
                        + correctCode.value());
                xsw.writeEndElement(); // End of <env:Value>
                xsw.writeEndElement(); // end of <env:Code>
            } else {   
                xsw.writeStartElement("faultcode");
                final String prefixEnv = xsw.getPrefix(SoapEnvelope.SOAP11_NS.value().toString());
                xsw.writeCharacters(
                        (prefixEnv == null || prefixEnv.length() == 0 ? "" : prefixEnv + ":" )
                        + correctCode.value());
                xsw.writeEndElement(); // end of <faultcode>
            }

            // create reason
            if (version.equals(BindingSoapVersion.SOAP12_NS)) {
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Reason");
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Text");
                xsw.writeAttribute("http://www.w3.org/XML/1998/namespace", "lang", "en");
                xsw.writeCharacters(reason);
                xsw.writeEndElement(); // End of <env:Text>
                xsw.writeEndElement(); // end of <env:Reason>
            } else {
                xsw.writeStartElement("faultstring");
                xsw.writeCharacters(reason);
                xsw.writeEndElement(); // end of <faultstring>
            }
            
            // create details
            if (version.equals(BindingSoapVersion.SOAP12_NS)) {
                xsw.writeStartElement(SoapEnvelope.SOAP12_NS.value().toString(), "Detail");
            } else {
                xsw.writeStartElement("detail");
            }
            for (Element element : details) {
                SOAPFault.writeElementToXMLStreamWriter(element, xsw);
            }
            xsw.writeEndElement(); // end of <env:Detail> or <detail>
            xsw.close();
            
            
            ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
            return new StreamSource(bais);
            
        } catch (final XMLStreamException e) {
            throw new PEtALSCDKException("Error: Impossible to detils in soap fault", e);
        }
    }
    
    private static void writeElementToXMLStreamWriter(final Element element, final XMLStreamWriter xsw) throws XMLStreamException {
        // Write the start tag
        {
            final String namespace = element.getNamespaceURI();
            if (namespace != null) {
                final String prefix = element.getPrefix();
                if (prefix != null) {
                    xsw.setPrefix(prefix, namespace);
                    xsw.writeStartElement(prefix, element.getLocalName(), namespace);
                    xsw.writeNamespace(prefix, namespace);
                }
                else {
                    xsw.writeStartElement(namespace, element.getLocalName());
                }
            } else {
                xsw.writeStartElement(element.getLocalName());
            }
        }
        
        // Write attributes
        final NamedNodeMap attrs = element.getAttributes();
        final int nbAttrs = attrs.getLength();
        for (int i=0; i<nbAttrs; i++) {
            final Node attrNode = attrs.item(i);
            final String attrNamespace = attrNode.getNamespaceURI();
            if (attrNamespace != null) {
                // We skip namespace declaration in element
                if (!"http://www.w3.org/2000/xmlns/".equals(attrNamespace)) {
                    final String attrPrefix = attrNode.getPrefix();
                    if (attrPrefix != null) {
                        xsw.setPrefix(attrPrefix, attrNamespace);
                        xsw.writeAttribute(attrPrefix, attrNode.getLocalName(), attrNamespace, attrNode.getNodeValue());
                    }
                    else {
                        xsw.writeAttribute(attrNamespace, attrNode.getLocalName(), attrNode.getNodeValue());
                    }
                }
            } else {
                xsw.writeAttribute(attrNode.getLocalName(), attrNode.getNodeValue());
            }
        }
        
        // Write Element content
        final String elementValue = element.getNodeValue();
        if (elementValue != null) {
            xsw.writeCharacters(elementValue);
        }
        
        // Write Sub-Elements & Content
        final NodeList childElts = element.getChildNodes();
        final int nbChildElt = childElts.getLength();
        for (int i=0; i<nbChildElt; i++) {
            final Node childEltNode = childElts.item(i);
            if (childEltNode instanceof Element) {
                SOAPFault.writeElementToXMLStreamWriter((Element)childEltNode, xsw);
            } else if (childEltNode instanceof Text) {
                xsw.writeCharacters(((Text)childEltNode).getData());
            }
        }
        
        // Write the end tag
        xsw.writeEndElement();
    }

}
